/*
create triggers on pgr_tables:

	function	|trigger
	----------------|------------------------
	pgr_auto_i      |pgr_autotrigger_i
	pgr_auto_u      |pgr_autotrigger_u
	pgr_auto_d      |pgr_autotrigger_d


when you add <tablename> to table pgr_tables, pgr_autotrigger_i creates replica triggers on 
table <tablename>, creates the auxiliary table <tablename_flow> and updates local pgr_distrib;
pgr_autotrigger_d checks if <tablename_flow> is empty and removes replica triggers on 
<tablename> and <tablename_flow>.
<tablename>_grid_idx index is created on <tablename>.pgr_grid in pgr_auto_i()
and dropped in pgr_auto_d()
*/



create function pgr_auto_i() returns opaque as '
elog NOTICE "working on $NEW(tablename) ..."
spi_exec "select getpgusername() as connected_user;"
if {[string compare $connected_user replicator]==0} {
spi_exec "select count(pgr_sid) as lsite_exist  from pgr_lsite"
set quotchar {"}
if { $lsite_exist ==0} {
	elog NOTICE "fill pgr_lsite first!"
	return SKIP
	}
spi_exec "select count(host) as hosts_exist  from pgr_hosts"
if { $hosts_exist ==0} {
	elog NOTICE "fill pgr_hosts first!"
	return SKIP
	}
spi_exec "select count(relname) as table_exists from pg_class where relname=\'[quote $NEW(tablename)]\'"
if { $table_exists == 0 } {
	elog NOTICE "$NEW(tablename) does not exist in this database!"
	return SKIP
	}
spi_exec "select count(*) as grid_exists from pg_attribute where attname =\'pgr_grid\' and attrelid = (select oid from pg_class where relname =\'[quote $NEW(tablename)]\')"
if { $grid_exists == 0 } {
	elog NOTICE "attribute \'pgr_grid\' missing in $NEW(tablename)!"
	return SKIP
	}
spi_exec "select count(relname) as flow_exists from pg_class where relname=\'[quote $NEW(tablename)_flow]\'"
spi_exec "select count(*) as pgrtriggers from pg_trigger where tgrelid = (select oid from pg_class where relname = \'$NEW(tablename)\') and tgname like \'pgr_%\'"
if { $pgrtriggers==0 } {
	spi_exec "create trigger pgr_insert before insert on ${quotchar}[quote $NEW(tablename)]${quotchar} for each row execute procedure pgr_insert()"
        spi_exec "create trigger pgr_update before update on ${quotchar}[quote $NEW(tablename)]${quotchar} for each row execute procedure pgr_update()"
        spi_exec "create trigger pgr_delete before delete on ${quotchar}[quote $NEW(tablename)]${quotchar} for each row execute procedure pgr_delete()"
       	elog NOTICE "Replica Triggers Created."
	} else {
	elog NOTICE "Triggers already exist"
	}
if { $flow_exists == 0 } {
	spi_exec "create table ${quotchar}[quote $NEW(tablename)_flow]${quotchar} as select * from pgr_flow_template a, ${quotchar}[quote $NEW(tablename)]${quotchar} b"
	spi_exec "revoke all on ${quotchar}[quote $NEW(tablename)_flow]${quotchar} from public"
	spi_exec "grant all on ${quotchar}[quote $NEW(tablename)_flow]${quotchar} to replicator"
	spi_exec "grant insert on ${quotchar}[quote $NEW(tablename)_flow]${quotchar} to public"
       	elog NOTICE "$NEW(tablename)_flow  Created"
	} else {
	elog NOTICE "$NEW(tablename)_flow already exist"
	}
spi_exec "insert into pgr_monitor (tablename, when_occurred) values (\'[quote $NEW(tablename)]\', \'now\')"
spi_exec "select pgr_sid as locale from pgr_lsite;"
spi_exec "insert into pgr_distrib  values (\'$locale\', \'[quote $NEW(tablename)]\')"
spi_exec "select count(*) as grid_idx_exists from pg_indexes where indexname = \'[quote $NEW(tablename)]_grid_idx\'"
if { $grid_idx_exists == 0 } {
	spi_exec "create unique index ${quotchar}[quote $NEW(tablename)]_grid_idx${quotchar} on ${quotchar}[quote $NEW(tablename)]${quotchar} (pgr_grid)"
	}
return [array get NEW]
} else {
elog NOTICE "user $connected_user cannot change this system table!"
return SKIP
}
' language 'pltcl';



create trigger pgr_autotrigger_i before insert on pgr_tables for each row execute procedure pgr_auto_i();




create function pgr_auto_u() returns opaque as '
spi_exec "select getpgusername() as connected_user;"
if {[string compare $connected_user replicator]==0} {
	if {[string compare $OLD(tablename) $NEW(tablename)]!=0} {
		elog NOTICE "You cannot update <tablename> attribute: try deleting the tuple and inserting a new one"
        	return SKIP
		} else {
		return OK
		}
} else {
elog NOTICE "user $connected_user cannot change this system table!"
return SKIP
}
' language 'pltcl';


create trigger pgr_autotrigger_u before update on pgr_tables for each row execute procedure pgr_auto_u();




create function pgr_auto_d() returns opaque as '
elog NOTICE "working on $OLD(tablename) ..."
spi_exec "select getpgusername() as connected_user;"
if {[string compare $connected_user replicator]==0} {
set quotchar {"}
spi_exec "select count(relname) as flow_exists from pg_class where relname=\'[quote $OLD(tablename)_flow]\'"
if { $flow_exists > 0 } {
	spi_exec "select count(*) as flows_present from ${quotchar}[quote $OLD(tablename)_flow]${quotchar}"
	if { $flows_present > 0} {
		elog NOTICE "$OLD(tablename)_flow must be empty: $flows_present tuples actually present "
		return SKIP
		}
	spi_exec "drop table ${quotchar}[quote $OLD(tablename)_flow]${quotchar}"
	elog NOTICE "$OLD(tablename)_flow dropped"
	}
spi_exec "drop trigger pgr_insert on ${quotchar}[quote $OLD(tablename)]${quotchar}"
spi_exec "drop trigger pgr_update on ${quotchar}[quote $OLD(tablename)]${quotchar}"
spi_exec "drop trigger pgr_delete on ${quotchar}[quote $OLD(tablename)]${quotchar}"
elog NOTICE "Replica triggers dropped on [quote $OLD(tablename)]"
spi_exec "delete from pgr_monitor where tablename = \'[quote $OLD(tablename)]\'"
spi_exec "select pgr_sid as locale from pgr_lsite;"
spi_exec "delete from pgr_distrib where tablename = \'[quote $OLD(tablename)]\' and site=$locale"
spi_exec "select count(*) as grid_idx_exists from pg_indexes where indexname = \'[quote $OLD(tablename)]_grid_idx\'"
if { $grid_idx_exists != 0 } {
	spi_exec "drop index ${quotchar}[quote $OLD(tablename)]_grid_idx${quotchar}"
	}
return OK
} else {
elog NOTICE "user $connected_user cannot change this system table!"
return SKIP
}
' language 'pltcl';


create trigger pgr_autotrigger_d before delete on pgr_tables for each row execute procedure pgr_auto_d();


