#!/bin/sh
# the next line restart using pgtclsh \
exec pgtclsh "$0" "$@"


# Usage: rollback.tcl <dbname> 

# Rollback users operations in local <dbname> getting
# the database state as it was after last replica session.
# a sql file (PGRDATA/rbase/<dbname>/log/dbname_rb.sql) 
# is produced to resubmit all operations.
# ATTENTION please:
# the sql file must be submitted as replicator,
# otherwhise pgr_grid will be modified during new insertion
# and subsequent operations related to the old pgr_grid will fail!!!

# Operations to roll back must have flow_flag=Y (the default in trigger capture) so we can
# roll back  operations according to some criteria (date, username, ...) by previously updating
# <table_flow> setting pgr_flow_flag=N to that rows we do not want to roll back.


proc init {} {
global pgrdVar 

        if {[catch {read_pgr.conf} msg]} {
                #puts "[clock clicks]  $msg"
                return -code error $msg
                }

        if {[catch {source [file join $pgrdVar(pgrd_HOME) setup.tcl]} msg]} {
                #puts "[clock clicks]  $msg"
                return -code error $msg
                }

        if {[catch {setup} msg]} {
                #puts stderr $msg
                return -code error $msg
                }

        # Loading all defined procedures modules
        foreach module { common pgproc  queryEngine} {
                source [file join $pgrdVar(pgrd_proc) $module.tcl]
                }
        }


proc read_pgr.conf {} {
global pgrdVar 
        if {[catch {open /etc/pgr.conf r } conf]} {
                #puts "[clock clicks] $conf"
                #puts stderr $conf
                return -code error $conf
                }

        while {![ eof $conf]} {
                gets $conf line
                #puts $line
                if {[string first # $line]!=0} {
                        if {[regexp {^([^= ]+)=([^ ]+)} $line match var value]} {
                                #puts "match: $match"
                                #puts "$var=$value"
                                set pgrdVar($var) $value
                                }
                        }
                }


        if {![info exists pgrdVar(PGRBIN)]} {
                puts stderr "No  PGRBIN  variable in /etc/pgr.conf"
                puts stderr "Is ReplicaTools properly installed ?"
                return -code error "No  PGRBIN  variable in /etc/pgr.conf"
                }

        set pgrdVar(pgrd_HOME) $pgrdVar(PGRBIN)

        }



proc pgrRI { dbname action } {
global pgrdVar
        set lhost $pgrdVar($dbname,lhost)

        set now "        "
        #puts  "$now [clock clicks] pgrRI $dbname $action IN"

        set tablelist {}
        foreach table $pgrdVar($dbname,$lhost,tables) {
                lappend tablelist \'[myquote $table]\'
                }
        set tablelistString [join $tablelist ", " ]

        if {[string compare $action "disable"]==0} {
                set enabled false
                set pgrdVar($dbname,RI) disabled
                set query "select oid from pg_trigger where tgname ~\'^RI_\' and tgrelid in (select oid from pg_class where relname in ($tablelistString))"
                #puts  "$now [clock clicks] RI: $query"
                set pgrdVar($dbname,RIoidsList) {}
                pg_select $pgrdVar($dbname,lconn) $query RIoid {
                        lappend pgrdVar($dbname,RIoidsList) $RIoid(oid)
                        }
        } else {
                set enabled true
                set pgrdVar($dbname,RI) enabled
        }

        if {[llength $pgrdVar($dbname,RIoidsList)]>0} {
                set RIoidsListString [join $pgrdVar($dbname,RIoidsList) ", "]
                set query "update pg_trigger set tgenabled=\'$enabled\' where oid in ($RIoidsListString)"
                #puts  "$now [clock clicks] RI: $query"
                set ris [query_exec $dbname $query]
                #puts  "$now [clock clicks] RI: $ris"
                } else {
                #puts "$now [clock clicks] RI: No Referential Integrity constraints on replicated tables."
                }


#puts "$now [clock clicks] pgrRI $dbname $action OUT"
}



#///////////
# MAIN
#///////////

if {[catch {init} errmsg ]} {
        puts stderr $errmsg
        exit
        }


puts [format "\npgReplicator Version: $pgrdVar(version)\nrollback utility\n"]

# only replicator can run this script
set uname [ exec id -un]
if {[string compare $uname replicator]!=0} {
        puts [format "Only user replicator can run this script!\n\a"]
        exit
        }


if  [expr (($argc < 1) || ($argc > 1))] {
	puts [format "Usage: rollback.tcl <dbname>\n\a"]
        exit
        }


set dbname [lindex $argv 0]
set quotchar {"}

# connect to dbname
if [catch {db_connect $dbname localhost $pgrdVar(pg_port)} err] {
	puts [format "Error connecting to $dbname:\n$err"]
	return -1
	} else {
	set lconn $pgrdVar($dbname,lconn)
	}



# check if dbname is a replicated database
set isreplicated 0
set query "select count(*) as isreplicated from pg_class where relname = \'pgr_tables\'"
set qry_handle [query_exec $dbname $query -return_handle]
pg_result $qry_handle -assign risp
set isreplicated $risp(0,isreplicated)
pg_result $qry_handle -clear
if { $isreplicated == 0} {
        puts [format "\n$dbname is not a replicated database!\n"]
	db_disconnect $dbname
	return -1
	}



# <dbname> replica directories
set pgrdVar($dbname,temp) [file join $pgrdVar(pgrd_DATA) rbase $dbname temp]
set pgrdVar($dbname,log) [file join $pgrdVar(pgrd_DATA) rbase $dbname log]
set pgrdVar($dbname,history) [file join $pgrdVar(pgrd_DATA) rbase $dbname history]
set pgrdVar($dbname,inbox) [file join $pgrdVar(pgrd_BOX) $dbname in]
set pgrdVar($dbname,outbox) [file join $pgrdVar(pgrd_BOX) $dbname out]


# calling distrib procedure;
# tables structure holds informations about table
# distribution among hosts for the specified database
if [catch {distrib $dbname} errmsg] {
        set now [clock format [clock seconds] -format %T]
	db_disconnect $dbname
        return -code error "ERR/distrib/$errmsg"
        }

set lhost  $pgrdVar($dbname,lhost)
set quotchar {"}

#foreach elem [array names pgrdVar] { puts "$elem $pgrdVar($elem)"}



set now [clock format [clock second] -format %T]
puts  "$now query_engine $dbname"


# disable RI
pgrRI $dbname disable


# open sql file
set sqlfname [file join $pgrdVar($dbname,temp) ${dbname}_rb.sql]
catch {file copy -force $sqlfname ${sqlfname}.old} ;# a secure copy
set sqlf [open  $sqlfname w+]
fconfigure $sqlf -buffering none
set pgrdVar($dbname,qesql) $sqlf

# only replicator can safetely resubmit operations stored in sql file!!!
puts $pgrdVar($dbname,qesql) "\\connect - replicator"

set flows_rb {}

# run query_engine to rollback user's tables
# and create sql file
set tablelist $pgrdVar($dbname,$lhost,tables)
foreach table $tablelist {

        # rollback local  operations 
        set now [clock format [clock second] -format %T]
        puts  "$now local users operations rollback: query_engine $dbname $table local passed back..."
        query_engine $dbname $table local passed back 
        set now [clock format [clock second] -format %T]
        puts  "$now ...Done"

        # create sql file with rolled back operations 
        set now [clock format [clock second] -format %T]
        puts  "$now local users operations rollback: query_engine $dbname $table local passed forward $sqlfname..."
        query_engine $dbname $table local passed forward $sqlfname
        set now [clock format [clock second] -format %T]
        puts  "$now ...Done"


	# transfer rolled back operations from table_flow to table_flow_rb
        set now [clock format [clock second] -format %T]
	set query "drop table  ${quotchar}${table}_flow_rb${quotchar}"
	puts "$now $query"
	catch {query_exec $dbname $query} ris
	puts "$now $ris"
	set query "select * into ${quotchar}${table}_flow_rb${quotchar} from ${quotchar}${table}_flow${quotchar} where pgr_flow_flag=\'Y\'"
	puts "$now $query"
	set ris [query_exec $dbname $query]
	puts "$now $ris"

	lappend $flows_rb ${table}_flow_rb

	# delete rolled back operations from table_flow
        set now [clock format [clock second] -format %T]
	set query "delete from ${quotchar}${table}_flow${quotchar} where pgr_flow_flag=\'Y\'"
	puts "$now $query"
	set ris [query_exec $dbname $query]
	puts "$now $ris"


        } ;# end of foreach table



# close sql file
close $sqlf


# reenable RI
pgrRI $dbname enable

puts [format "\nthe file:\n$sqlfname\n was produced to locally resubmit rolled back operations (as user replicator!)\n"]

if {[llength $flows_rb]>0} {
	puts [format "auxiliary tables:\n$flows_rb\n were populated with  rolled back operations \n"]
	}

# disconnect from dbname
db_disconnect $dbname

puts " "

